// RTC-Treiber   Dr. Weisser 13.10.2016     20.10.2019
//***************************************************************************
//  File........: ds3231.c
//  Target(s)...: ATmega328p
//  Compiler....: AVR-GCC
//  Description.: RTC routines for real time clock
//                  DS1302: SPI-Schnittstelle 500kHz-2MHz
//                  DS3231: I2C, Quarz, 2ppm, Gangkorrektur, Temp.sensor 
//                  DS1307: I2C, 56byte RAM, 4.5-5.5V, Ibat=300-500nA 
//                  PCF8563: I2C, 1-5.5V, Ibat=250nA
//                  RV3028: I2C, Quarz, 5ppm, 2byte RAM, 1.1-5.5V, 40nA
//
//                Vorsicht: DS3231 braucht bis 300ms nach PowerUp
//                um brauchbar zu arbeiten. Ggf. RST auswerten
//                oder RST mit Reset des uC verbinden. 
//***************************************************************************
/* 13.10.2016 Uebernahme RTC-Code von ds1302.c
   14.10.2016 mit #define DS1302 Umschaltung DS1302 und DS3231
   15.10.2016 ueberarbeitet. delay_us(1) statt 2us. Test dazu noetig.
              bei RTC_set_time_date() wday noch eingebaut
              wday statt day macht es klarer
              Code optimiert fuer RTC_read()
              Namen veraendert fuer einfacheres Verstaendnis
   16.10.2016 Auslesen Sekunden klappt nicht. kommt immer 0.
              Ansprechen der Pins ist ok. Ueberschwinger zu sehen
              Debug-Befehle in Code eingebaut. #define DEBUG 1 hinzu
              Fehler in RTC_read() behoben
              RTC_read() unterscheidet nun Lesen RAM und Lesen Clock
              RTC_read_byte() liest nun nur "byte" und manipuliert dies
              neue Funktion RTC_read_clock() zeigt berechnete Clock-Werte
              1us delay statt 2us wieder eingebaut. Geht so !
   17.10.2016 Problem dass Uhr blockiert wird und so 2x hintereinander
              die Uhr mit sec=10 ausgelesen wird
              RST systematisch am Funktionsende wieder auf low setzen:
              - in RTC_get_fattime(void)
              - in RTC_read_burst(void)
              bei ReadBurst wurde nicht alles gelesen -> die Uhr
              wartet dann auf Abschluss des Lesens. der kommt nicht.
              die Uhr kann daher die Register nicht aktualisieren.
              -> daher wird der falsche alte Wert ausgelesen.
              geht nun.
              defines fuer DS3231-Register gemacht
              define-Namen alle GROSS nun
   18.10.2016 in RTC_read() I2C-Abfrage eingebaut fuer ds3231.c
              RTC_set_time_date korrigiert. Hex-Eingabe dort noetig !
              RTC_read(WDAY) liefert 2 (korrekt) und
              RTC_read_burst() liefert 6 (falsch) stattdessen. 
              Fehler korrigiert. 435kHz fuer I2C klappt 
    5.12.2016 USART-Ausgaben an neue usart.c angepasst 
              i2c_write -> i2c_write_block 
              i2c_read  -> i2c_read_block 
              RTC_RST_LOW  -> RTC_RST_L 
              RTC_RST_HIGH -> RTC_RST_H 
              Test DS1302 ok. 
              Test DS3231 ohne TWI ok. Stellen Uhr ok. 
    6.12.2016 neue Funktion out_u08_2 fuer Zeit-/Datum-Ausgabe 
    7.12.2016 i2c_read_block geht nicht mit TWI 
              liegt an Status-Rueckmeldung 
              Fehler behoben 
   29.12.2016 defines fuer PCF8563 
    24.2.2019 defines fuer Pins DS1302 nach ds3231.h 
              ueberarbeitet Pin-Definition 
              #define RTC_IO_DIN -> #define RTC_IO_VAL 
    26.2.2019 rtc_test() hinzu 
              Funktionsnamen rtc_xxxx() statt RTC_xxx() 
    28.2.2019 RV3028 hinzu 
     1.3.2019 rtc_lcd() fuer Ausgabe LCD hinzu 
   20.10.2019 Debug-Ausgaben outs(USART,PSTR) -> outs_p(USART,PSTR) 

*/

#include <stdint.h>       // Integer-Datentypen 8-64bit
#include <avr/io.h>       // Pins und Interrupt-Tabelle -> siehe auch iom328p.h
#include "main.h"         // CPU-Frequenz, sbi, cbi
#include "delay.h"        // Wartezeiten 2us, ms
#include "ds3231.h"       // RTC-Pins, RTC-Funktionen + Prototypen
#include "i2c.h"          // I2C-Zugriffe
#include "usart.h"        // USART-Funktionen fuer Debug-Ausgaben
#include "out.h"          // Ausgaben USART, LCD
#include <avr/pgmspace.h> // PSTR

#define DS1302  0    // 1: DS1302         0: kein DS1302 siehe ds3231.h
#define DS3231  1    // 1: DS3231/DS1307  0: kein DS3231/DS1307 
#define PCF8563 0    // 1: PCF8563        0: kein PCF8563 
#define RV3028  0    // 1: RV3028         0: kein RV3028 

#define DEBUG  0     // 0: keine Debug-Ausgaben 1: Debug-Ausgaben
#define DEBUG1 0     // 0: keine Debug-Ausgaben 1: Debug-Ausgaben


// set RTC to time, date
void rtc_set_time_date(unsigned char date,  // 0x31 31.12.15 20:17:00
                       unsigned char month, // 0x12
                       unsigned char year,  // 0x15
                       unsigned char hour,  // 0x20
                       unsigned char min,   // 0x17
                       unsigned char sec,   // 0x00
                       unsigned char wday)  // 0x02
{
   // RTC setzen: Datum, Uhrzeit. noetig nach Batteriewechsel

   // beachte: Hexadezimal-Eingabe !!
   // Sekunden z.B. 59 muessen in 2 Halbbytes: 5 und 9
   // 59 / 10 = 5 + 9
   //        -> 5 ins vordere Halbbyte (*16)
   //        -> 9 ins hintere Halbbyte
   // *16 = *2 *2 *2 *2 = 4x shift links
   // *16 = <<4
   // Hex-Eingabe leistet das ohne die folgende Umrechung


/* ggf. Eingaben umrechnen von dezimal in hex:
   diesen Code kann man ggf. sparen.
   diese Funktion braucht es ja sehr selten.
   outs_p(USART,PSTR("\nRTC set: \nsec: ")); // \nRTC: ausgeben
   out_u08(USART,sec,2,0);         // ausgeben
   temp = sec / 10 * 16 + sec % 10;
   outs_p(USART,PSTR(" temp: "));
   out_u08(USART,temp,2,0);        // ausgeben

   outs_p(USART,PSTR("\nmin: "));
   out_u08(USART,min,2,0);         // ausgeben
   temp = min / 10 * 16 + min % 10;
   outs_p(USART,PSTR(" temp: "));
   out_u08(USART,temp,2,0);        // ausgeben

   outs_p(USART,PSTR("\nhour: "));
   out_u08(USART,hour,2,0);        // ausgeben
   temp = hour / 10 * 16 + hour % 10;
   outs_p(USART,PSTR(" temp: "));
   out_u08(USART,temp,2,0);        // ausgeben

   outs_p(USART,PSTR("\nwday: "));
   out_u08(USART,wday,1,0);        // ausgeben
   temp = wday & 7;  // nur untere 3 bit schreiben
   outs_p(USART,PSTR(" temp: "));
   out_u08(USART,temp,1,0);        // ausgeben

   outs_p(USART,PSTR("\ndate: "));
   out_u08(USART,date,2,0);        // ausgeben
   temp = date / 10 * 16 + date % 10;
   outs_p(USART,PSTR(" temp: "));
   out_u08(USART,temp,2,0);        // ausgeben

   outs_p(USART,PSTR("\nmonth: "));
   out_u08(USART,month,2,0);       // ausgeben
   temp = month / 10 * 16 + month % 10;
   outs_p(USART,PSTR(" temp: "));
   out_u08(USART,temp,2,0);        // ausgeben

   outs_p(USART,PSTR("\nyear: "));
   out_u08(USART,year,2,0);        // ausgeben
   temp = year / 10 * 16 + year % 10;
   outs_p(USART,PSTR(" temp: "));
   out_u08(USART,temp,2,0);        // ausgeben
*/


#if DS1302
   rtc_write(W_PROTECT,0x00); // RTC-Schreibschutz aufheben, falls gesetzt

   rtc_write(YEAR_W,year);    // Jahr 2016=16
   rtc_write(MONTH_W,month);  // Monat Sept=09
   rtc_write(DATE_W,date);    // Datum 4.=04
   rtc_write(HOUR_W,hour);    // Stunde 21=21
   rtc_write(MIN_W,min);      // Minuten 20=20
   rtc_write(SEC_W,sec);      // Sekunden 0=00
   rtc_write(WDAY_W,wday);    // Wochentag Samstag=SAT

   // Beispiel:
   //rtc_write(YEAR_W,0x16);  // Jahr 2016=16
   //rtc_write(MONTH_W,0x09); // Monat Sept=09
   //rtc_write(DATE_W,0x10);  // Datum 10.=10
   //rtc_write(HOUR_W,0x03);  // Stunde 3=03
   //rtc_write(MIN_W,0x44);   // Minuten 44=44
   //rtc_write(SEC_W,0x00);   // Sekunden 00=00
   //rtc_write(WDAY_W,SAT);   // Wochentag Samstag=SAT
#endif // DS1302

#if DS3231 // DS3231/DS1307
   unsigned char buf[8]; // Puffer fuellen zum blockweisen schreiben

   buf[0] = sec;   // 0-59
   buf[1] = min;   // 0-59
   buf[2] = hour;  // 0-23
   buf[3] = wday;  // 1-7
   buf[4] = date;  // 1-31
   buf[5] = month; // 1-12
   buf[6] = year;  // 0-99

   // RTC ueber I2C schreiben - blockweise
   // schreibt 7 byte ab Adresse 0x00: (Datenblatt S.11)
   //   - Sekunden
   //   - Minuten
   //   - Stunden
   //   - Wochentag (die Uhr sollte das selbst koennen)
   //   - Tag
   //   - Monat (Vorsicht: ganz vorne steht Jahrhundert-bit)
   //   - Jahr
   i2c_write_block(RTC_ADR_W, 0, 7, buf); // Routine von Chan
                // unsigned char dev               device address
                //            unsigned char adr    write start address
                //               unsigned int cnt  write byte count
                //                  puffer         data to be written
#endif // DS3231/DS1307

#if PCF8563 // PCF8563
   unsigned char buf[8]; // Puffer fuellen zum blockweisen schreiben

   buf[0] = sec;    // 0-59
   buf[1] = min;    // 0-59
   buf[2] = hour;   // 0-23
   buf[3] = date;   // 1-31
   buf[4] = wday-1; // 1-7 -> 0-6 
   buf[5] = month;  // 1-12
   buf[6] = year;   // 0-99

   // RTC ueber I2C schreiben - blockweise
   // schreibt 7 byte ab Adresse 0x02: (Datenblatt S.6)
   //   - Sekunden (Vorsicht: ganz vorne steht VL)
   //   - Minuten
   //   - Stunden
   //   - Tag im Monat
   //   - Wochentag
   //   - Monat (Vorsicht: ganz vorne steht Jahrhundert-bit)
   //   - Jahr
   i2c_write_block(RTC_ADR_W, 2, 7, buf); // Routine von Chan
                // unsigned char dev               device address
                //            unsigned char adr    write start address
                //               unsigned int cnt  write byte count
                //                  puffer         data to be written
#endif // PCF8563

#if RV3028 // RV3028
   unsigned char buf[8]; // Puffer fuellen zum blockweisen schreiben

   buf[0] = sec;    // 0-59
   buf[1] = min;    // 0-59
   buf[2] = hour;   // 0-23
   buf[3] = wday-1; // 1-7 -> 0-6 
   buf[4] = date;   // 1-31
   buf[5] = month;  // 1-12
   buf[6] = year;   // 0-99

   // RTC ueber I2C schreiben - blockweise
   // schreibt 7 byte ab Adresse 0x00: (Datenblatt S.14)
   //   - Sekunden
   //   - Minuten
   //   - Stunden
   //   - Wochentag
   //   - Tag im Monat
   //   - Monat
   //   - Jahr
   i2c_write_block(RTC_ADR_W, 0, 7, buf); // Routine von Chan
                // unsigned char dev               device address
                //            unsigned char adr    write start address
                //               unsigned int cnt  write byte count
                //                  puffer         data to be written
#endif // RV3028
}



// function for FatFs module
unsigned long rtc_get_fattime(void) // RTC im Burst lesen und packen fuer SD-Karte
{
   unsigned char sec,min,hour,date,month,wday,year;

#if DS1302
   //rtc_init();             // definierter Zustand der RTC-Ansteuerpins
   //RTC_RST_L();            // RST/CE auf low. sollte lo sein
   RTC_RST_H();              // RST/CE auf high

   rtc_write_byte(C_BURST);  // Register schreiben fuer RTC-Auslesen mittels Burst

   //outs_p(USART,PSTR("\nRTC-Fattime: ")); // \nRTC: ausgeben
   sec   = rtc_read_clock(); // Auslesen Sekunden 0-59 in Variable
   min   = rtc_read_clock(); // Auslesen Minuten 0-59 in Variable
   hour  = rtc_read_clock(); // Auslesen Stunden 0-23 in Variable
   date  = rtc_read_clock(); // Auslesen Tage 1-31 in Variable
   month = rtc_read_clock(); // Auslesen Monat 1-12 in Variable
   wday  = rtc_read_clock(); // Auslesen Wochentag 1-7 in Variable
   year  = rtc_read_clock(); // Auslesen Jahr 0-99 in Variable

   RTC_RST_L();              // RST/CE auf low
   //RTC_CLK_L();            // SCLK auf low. sollte lo sein
#endif // DS1302

#if DS3231 // DS3231/DS1307
   unsigned char buf[8];     // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen - blockweise
   // liest 7 byte ab Adresse 0x00: (Datenblatt S.11)
   //   - Sekunden
   //   - Minuten
   //   - Stunden
   //   - Wochentag
   //   - Tag im Monat
   //   - Monat (Vorsicht: ganz vorne steht Jahrhundert-bit)
   //   - Jahr
   i2c_read_block(RTC_ADR_W, 0, 7, buf); // Routine von Chan
               // unsigned char dev               device address
               //            unsigned char adr    write start address
               //               unsigned int cnt  write byte count
               //                  puffer         data to be written

   // Umrechnung BCD -> binaer
   sec  = (buf[0] & 0x0F) + ((buf[0] >> 4) & 7) * 10; // 0-59
   min  = (buf[1] & 0x0F) +  (buf[1] >> 4) * 10;      // 0-59
   hour = (buf[2] & 0x0F) + ((buf[2] >> 4) & 3) * 10; // 0-23
   wday = (buf[3] & 0x07);                            // 1-7
   date = (buf[4] & 0x0F) + ((buf[4] >> 4) & 3) * 10; // 1-31
   month= (buf[5] & 0x0F) + ((buf[5] >> 4) & 1) * 10; // 1-12
   year = (buf[6] & 0x0F) +  (buf[6] >> 4) * 10;      // 0-99
#endif // DS3231/DS1307

#if PCF8563 // PCF8563
   unsigned char buf[8];     // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen - blockweise
   // liest 7 byte ab Adresse 0x02: (Datenblatt S.6)
   //   - Sekunden (Vorsicht: ganz vorne steht VL)
   //   - Minuten
   //   - Stunden
   //   - Tag im Monat
   //   - Wochentag
   //   - Monat (Vorsicht: ganz vorne steht Jahrhundert-bit)
   //   - Jahr
   i2c_read_block(RTC_ADR_W, 2, 7, buf); // Routine von Chan
               // unsigned char dev               device address
               //            unsigned char adr    write start address
               //               unsigned int cnt  write byte count
               //                  puffer         data to be written

   // Umrechnung BCD -> binaer
   sec  = (buf[0] & 0x0F) + ((buf[0] >> 4) & 7) * 10; // 0-59
   min  = (buf[1] & 0x0F) +  (buf[1] >> 4) * 10;      // 0-59
   hour = (buf[2] & 0x0F) + ((buf[2] >> 4) & 3) * 10; // 0-23
   date = (buf[3] & 0x0F) + ((buf[3] >> 4) & 3) * 10; // 1-31
   wday = (buf[4] & 0x07) + 1;                        // 0-6 -> 1-7
   month= (buf[5] & 0x0F) + ((buf[5] >> 4) & 1) * 10; // 1-12
   year = (buf[6] & 0x0F) +  (buf[6] >> 4) * 10;      // 0-99
#endif // PCF8563

#if RV3028 // RV3028
   unsigned char buf[8];     // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen - blockweise
   // liest 7 byte ab Adresse 0x00: (Datenblatt S.14)
   //   - Sekunden
   //   - Minuten
   //   - Stunden
   //   - Wochentag
   //   - Tag im Monat
   //   - Monat
   //   - Jahr
   i2c_read_block(RTC_ADR_W, 0, 7, buf); // Routine von Chan
               // unsigned char dev               device address
               //            unsigned char adr    write start address
               //               unsigned int cnt  write byte count
               //                  puffer         data to be written

   // Umrechnung BCD -> binaer
   sec  = (buf[0] & 0x0F) +  (buf[0] >> 4) * 10;      // 0-59
   min  = (buf[1] & 0x0F) +  (buf[1] >> 4) * 10;      // 0-59
   hour = (buf[2] & 0x0F) +  (buf[2] >> 4) * 10;      // 0-23
   wday = (buf[3] & 0x07) + 1;                        // 0-6 -> 1-7
   date = (buf[4] & 0x0F) +  (buf[3] >> 4) * 10;      // 1-31
   month= (buf[5] & 0x0F) +  (buf[5] >> 4) * 10;      // 1-12
   year = (buf[6] & 0x0F) +  (buf[6] >> 4) * 10;      // 0-99
#endif // RV3028

   // pack date and time into 4 byte DWORD variable
   // xxxx xxxx xxxx xxxx xxxx xxxx xxxx xxxx  Variable 32bit
   // xxxx xxx                                 Jahr 0-127 ab 1980
   //         x xxx                            Monat 1-12
   //              x xxxx                      Tag 1-31
   //                     xxxx x               Stunde 0-24
   //                           xxx xxx        Minute 0-59
   //                                  x xxxx x  Sekunde 0-59 Sekunde fehlt !
   return  ((unsigned long)(2000 + year - 1980) << 25)
         | ((unsigned long)month << 21)
         | ((unsigned long)date << 16)
         | ((unsigned long)hour << 11)
         | ((unsigned long)min << 5)
         | ((unsigned long)sec >> 1);
}



void rtc_read_burst(void)    // alle 7 bytes der RTC rasch lesen/ausgeben
{
   unsigned char sec,min,hour,date,month,wday,year;

#if DS1302
   //rtc_init();             // definierter Zustand RTC-Ansteuerpins
   //RTC_RST_L();            // RST/CE auf low. sollte lo sein
   RTC_RST_H();              // RST/CE auf high

   rtc_write_byte(C_BURST);  // Register schreiben fuer Auslesen mittels Burst

   sec   = rtc_read_clock(); // Auslesen Sekunden 0-59 in Variable
   min   = rtc_read_clock(); // Auslesen Minuten 0-59 in Variable
   hour  = rtc_read_clock(); // Auslesen Stunden 0-23 in Variable
   date  = rtc_read_clock(); // Auslesen Tage 1-31 in Variable
   month = rtc_read_clock(); // Auslesen Monat 1-12 in Variable
   wday  = rtc_read_clock(); // Auslesen Wochentag 1-7 in Variable
   year  = rtc_read_clock(); // Auslesen Jahr 0-99 in Variable

   RTC_RST_L();              // RST/CE auf low
   //RTC_CLK_L();            // SCLK auf low. sollte lo sein
#endif // DS1302

#if DS3231 // DS3231/DS1307
   unsigned char buf[8];     // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen - blockweise
   // liest 7 byte ab Adresse 0x00: (Datenblatt S.11)
   //   - Sekunden
   //   - Minuten
   //   - Stunden
   //   - Wochentag (die Uhr sollte das selbst koennen)
   //   - Tag
   //   - Monat (Vorsicht: ganz vorne steht Jahrhundert-bit)
   //   - Jahr
   i2c_read_block(RTC_ADR_W, 0, 7, buf); // Routine von Chan
               // unsigned char dev               device address
               //            unsigned char adr    write start address
               //               unsigned int cnt  write byte count
               //                  puffer         data to be written

   // Umrechnung BCD -> binaer
   sec  = (buf[0] & 0x0F) + ((buf[0] >> 4) & 7) * 10; // 0-59
   min  = (buf[1] & 0x0F) +  (buf[1] >> 4) * 10;      // 0-59
   hour = (buf[2] & 0x0F) + ((buf[2] >> 4) & 3) * 10; // 0-23
   wday = (buf[3] & 0x07);                            // 1-7
   date = (buf[4] & 0x0F) + ((buf[4] >> 4) & 3) * 10; // 1-31
   month= (buf[5] & 0x0F) + ((buf[5] >> 4) & 1) * 10; // 1-12
   year = (buf[6] & 0x0F) +  (buf[6] >> 4) * 10;      // 0-99
#endif // DS3231/DS1307

#if PCF8563 // PCF8563
   unsigned char buf[8];     // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen - blockweise
   // liest 7 byte ab Adresse 0x02: (Datenblatt S.6)
   //   - Sekunden (Vorsicht: ganz vorne steht VL)
   //   - Minuten
   //   - Stunden
   //   - Tag im Monat
   //   - Wochentag
   //   - Monat (Vorsicht: ganz vorne steht Jahrhundert-bit)
   //   - Jahr
   i2c_read_block(RTC_ADR_W, 2, 7, buf); // Routine von Chan
               // unsigned char dev               device address
               //            unsigned char adr    write start address
               //               unsigned int cnt  write byte count
               //                  puffer         data to be written

   // Umrechnung BCD -> binaer
   sec  = (buf[0] & 0x0F) + ((buf[0] >> 4) & 7) * 10; // 0-59
   min  = (buf[1] & 0x0F) +  (buf[1] >> 4) * 10;      // 0-59
   hour = (buf[2] & 0x0F) + ((buf[2] >> 4) & 3) * 10; // 0-23
   date = (buf[3] & 0x0F) + ((buf[3] >> 4) & 3) * 10; // 1-31
   wday = (buf[4] & 0x07) + 1;                        // 0-6 -> 1-7
   month= (buf[5] & 0x0F) + ((buf[5] >> 4) & 1) * 10; // 1-12
   year = (buf[6] & 0x0F) +  (buf[6] >> 4) * 10;      // 0-99
#endif // PCF8563

#if RV3028 // RV3028
   unsigned char buf[8];     // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen - blockweise
   // liest 7 byte ab Adresse 0x00: (Datenblatt S.14)
   //   - Sekunden
   //   - Minuten
   //   - Stunden
   //   - Wochentag
   //   - Tag im Monat
   //   - Monat
   //   - Jahr
   i2c_read_block(RTC_ADR_W, 0, 7, buf); // Routine von Chan
               // unsigned char dev               device address
               //            unsigned char adr    write start address
               //               unsigned int cnt  write byte count
               //                  puffer         data to be written

   // Umrechnung BCD -> binaer
   sec  = (buf[0] & 0x0F) +  (buf[0] >> 4) * 10;      // 0-59
   min  = (buf[1] & 0x0F) +  (buf[1] >> 4) * 10;      // 0-59
   hour = (buf[2] & 0x0F) +  (buf[2] >> 4) * 10;      // 0-23
   wday = (buf[3] & 0x07) + 1;                        // 0-6 -> 1-7
   date = (buf[4] & 0x0F) +  (buf[3] >> 4) * 10;      // 1-31
   month= (buf[5] & 0x0F) +  (buf[5] >> 4) * 10;      // 1-12
   year = (buf[6] & 0x0F) +  (buf[6] >> 4) * 10;      // 0-99
#endif // RV3028

   outs_p(USART,PSTR("\nRTC-Burst: ")); // \nRTC: ausgeben
   out_u08(USART,date,2,0);   // Datum ausgeben
   outc(USART,'.');           // . ausgeben
   out_u08_2(USART,month);    // Monat ausgeben
   outc(USART,'.');           // . ausgeben
   out_u08_2(USART,year);     // Jahr ausgeben
   outc(USART,' ');           // Space ausgeben
   out_u08(USART,hour,2,0);   // Stunde ausgeben
   outc(USART,':');           // : ausgeben
   out_u08_2(USART,min);      // Minuten ausgeben
   outc(USART,':');           // : ausgeben
   out_u08_2(USART,sec);      // Sekunden ausgeben
   outs_p(USART,PSTR(" Tag: "));
   out_u08(USART,wday,1,0);   // Wochentag als Zahl ausgeben
   outc(USART,'\n');          // neue Zeile ausgeben
}



unsigned char rtc_read(unsigned char reg) // read byte with register reg
// auslesen Datum-Tag:   byte= rtc_read(DATE);  Datum    3.=0x03
//          Wochentag:   byte= rtc_read(WDAY);  Tag     SAT=0x06
//          Datum-Monat: byte= rtc_read(MONTH); Monat  Sept=0x09
//          Datum-Jahr:  byte= rtc_read(YEAR);  Jahr   2016=0x16
//          Zeit-Stunde: byte= rtc_read(HOUR);  Stunde   11=0x11
//          Zeit-Minute: byte= rtc_read(MIN);   Minuten  40=0x40
//          Zeit-Sekunde:byte= rtc_read(SEC);   Sekunden 00=0x00
//          RAM:         byte= rtc_read(0b11xx xxx1);
{
   unsigned char byte;

#if DS1302
   //rtc_init();         // definierter Zustand der Ansteuerpins
   //RTC_RST_L();        // RST/CE auf low. sollte lo sein
   RTC_RST_H();          // RST/CE auf high

   rtc_write_byte(reg);  // Register schreiben fuer Auslesen

   // unterscheide Auslesen RAM und Clock
   // + bei RAM alle bits so lassen wie sie sind
   // + bei Clock vordere 4 bits * 10 + hintere 4 bits rechnen
   if ((reg & 0xc0)==0xc0){   // 11xx xxx1 zeigt RAM an, sonst Clock
      byte = rtc_read_byte(); // Auslesen roh in Variable

#if DEBUG
      outs_p(USART,PSTR("\nRAM-Zugriff: byte= ")); // Debug-Ausgabe
      usart_putc_hex(byte);           // byte als hex ausgeben
#endif // DEBUG

   }
   else {
      byte = rtc_read_clock(); // Auslesen in Variable + Korrektur fuer Clock

#if DEBUG
      outs_p(USART,PSTR("\nClock-Zugriff: byte= ")); // Debug-Ausgabe
      out_u08(USART,byte,2,0);           // byte ausgeben
#endif // DEBUG

   }

   RTC_RST_L();          // RST/CE auf low
   //RTC_CLK_L();        // SCLK auf low. sollte lo sein

#endif // DS1302

#if DS3231 // DS3231/DS1307
   unsigned char buf[2];  // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen 
   switch (reg){  // unterscheide Auswertung je nach Register
   case DATE:  // Datum auslesen 1-31
      i2c_read_block(RTC_ADR_W, DATE, 1, buf); // DATE abfragen
      byte = (buf[0] & 0x0F) + ((buf[0] >> 4) & 3) * 10; // 1-31
      break;
   case WDAY:  // Wochentag auslesen 1-7
      i2c_read_block(RTC_ADR_W, WDAY, 1, buf); // WDAY abfragen
      byte = (buf[0] & 0x07);                            // 1-7
      break;
   case MONTH: // Monat auslesen 1-12
      i2c_read_block(RTC_ADR_W, MONTH, 1, buf); // MONTH abfragen
      byte = (buf[0] & 0x0F) + ((buf[0] >> 4) & 1) * 10; // 1-12
      break;
   case YEAR:  // Jahr auslesen 0-99 (z.B. 16 fuer 2016)
      i2c_read_block(RTC_ADR_W, YEAR, 1, buf); // YEAR abfragen
      byte = (buf[0] & 0x0F) +  (buf[0] >> 4) * 10;      // 0-99
      break;
   case HOUR:  // Stunde auslesen 0-23
      i2c_read_block(RTC_ADR_W, HOUR, 1, buf); // HOUR abfragen
      byte = (buf[0] & 0x0F) + ((buf[0] >> 4) & 3) * 10; // 0-23
      break;
   case MIN:   // Minuten auslesen 0-59
      i2c_read_block(RTC_ADR_W, MIN, 1, buf); // MIN abfragen
      byte = (buf[0] & 0x0F) +  (buf[0] >> 4) * 10;      // 0-59
      break;
   case SEC:   // Sekunden auslesen 0-59
      i2c_read_block(RTC_ADR_W, SEC, 1, buf); // SEC abfragen
      byte = (buf[0] & 0x0F) + ((buf[0] >> 4) & 7) * 10; // 0-59
      break;
   default:    // sollte nie betreten werden
      byte = 99;    // bei Fehler dies ausgeben -> gut zu sehen !
   }
#endif // DS3231/DS1307

#if PCF8563 // PCF8563
   unsigned char buf[2];  // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen 
   switch (reg){  // unterscheide Auswertung je nach Register
   case DATE:  // Datum auslesen 1-31
      i2c_read_block(RTC_ADR_W, DATE, 1, buf); // DATE abfragen
      byte = (buf[0] & 0x0F) + ((buf[0] >> 4) & 3) * 10; // 1-31
      break;
   case WDAY:  // Wochentag auslesen 1-7
      i2c_read_block(RTC_ADR_W, WDAY, 1, buf); // WDAY abfragen
      byte = (buf[0] & 0x07)+1;                          // 0-6 -> 1-7
      break;
   case MONTH: // Monat auslesen 1-12
      i2c_read_block(RTC_ADR_W, MONTH, 1, buf); // MONTH abfragen
      byte = (buf[0] & 0x0F) + ((buf[0] >> 4) & 1) * 10; // 1-12
      break;
   case YEAR:  // Jahr auslesen 0-99 (z.B. 16 fuer 2016)
      i2c_read_block(RTC_ADR_W, YEAR, 1, buf); // YEAR abfragen
      byte = (buf[0] & 0x0F) +  (buf[0] >> 4) * 10;      // 0-99
      break;
   case HOUR:  // Stunde auslesen 0-23
      i2c_read_block(RTC_ADR_W, HOUR, 1, buf); // HOUR abfragen
      byte = (buf[0] & 0x0F) + ((buf[0] >> 4) & 3) * 10; // 0-23
      break;
   case MIN:   // Minuten auslesen 0-59
      i2c_read_block(RTC_ADR_W, MIN, 1, buf); // MIN abfragen
      byte = (buf[0] & 0x0F) +  (buf[0] >> 4) * 10;      // 0-59
      break;
   case SEC:   // Sekunden auslesen 0-59
      i2c_read_block(RTC_ADR_W, SEC, 1, buf); // SEC abfragen
      byte = (buf[0] & 0x0F) + ((buf[0] >> 4) & 7) * 10; // 0-59
      break;
   default:    // sollte nie betreten werden
      byte = 99;    // bei Fehler dies ausgeben -> gut zu sehen !
   }
#endif // PCF8563

#if RV3028 // RV3028
   unsigned char buf[2];  // Puffer fuer Lesen der RTC

   // RTC ueber I2C lesen 
   switch (reg){  // unterscheide Auswertung je nach Register
   case DATE:  // Datum auslesen 1-31
      i2c_read_block(RTC_ADR_W, DATE, 1, buf); // DATE abfragen
      byte = (buf[0] & 0x0F) + (buf[0] >> 4) * 10;       // 1-31
      break;
   case WDAY:  // Wochentag auslesen 1-7
      i2c_read_block(RTC_ADR_W, WDAY, 1, buf); // WDAY abfragen
      byte = buf[0]+1;                                   // 0-6 -> 1-7
      break;
   case MONTH: // Monat auslesen 1-12
      i2c_read_block(RTC_ADR_W, MONTH, 1, buf); // MONTH abfragen
      byte = (buf[0] & 0x0F) + (buf[0] >> 4) * 10;       // 1-12
      break;
   case YEAR:  // Jahr auslesen 0-99 (z.B. 16 fuer 2016)
      i2c_read_block(RTC_ADR_W, YEAR, 1, buf); // YEAR abfragen
      byte = (buf[0] & 0x0F) + (buf[0] >> 4) * 10;       // 0-99
      break;
   case HOUR:  // Stunde auslesen 0-23
      i2c_read_block(RTC_ADR_W, HOUR, 1, buf); // HOUR abfragen
      byte = (buf[0] & 0x0F) + (buf[0] >> 4) * 10;       // 0-23
      break;
   case MIN:   // Minuten auslesen 0-59
      i2c_read_block(RTC_ADR_W, MIN, 1, buf); // MIN abfragen
      byte = (buf[0] & 0x0F) + (buf[0] >> 4) * 10;       // 0-59
      break;
   case SEC:   // Sekunden auslesen 0-59
      i2c_read_block(RTC_ADR_W, SEC, 1, buf); // SEC abfragen
      byte = (buf[0] & 0x0F) + (buf[0] >> 4) * 10;       // 0-59
      break;
   default:    // sollte nie betreten werden
      byte = 99;    // bei Fehler dies ausgeben -> gut zu sehen !
   }
#endif // RV3028

   return byte;
}



#if DS1302
void rtc_write(unsigned char reg, unsigned char data) // write data with register reg
{
   //rtc_init();         // definierter Zustand der Ansteuerpins
   //RTC_RST_L();        // RST/CE auf low. sollte lo sein
   RTC_RST_H();          // RST/CE auf high

   rtc_write_byte(reg);  // erst Register schreiben
   rtc_write_byte(data); // dann Daten dazu schreiben lt. Datenblatt

   RTC_RST_L();          // RST/CE auf low
   //RTC_CLK_L();        // SCLK auf low. sollte lo sein
}


/*
void rtc_init(void)   // sets pins to begin and end ds1302 communication
{
   //RTC_RST_OUT();      // RST/CE-Pin auf Ausgang. sollte in main.c geschehen
   //RTC_CLK_OUT();      // CLK-Pin auf Ausgang. sollte in main.c geschehen
   //RTC_CLK_L();        // SCLK auf low. sollte in main.c geschehen
   RTC_RST_L();          // RST/CE auf low. sollte in main.c geschehen
   RTC_RST_H();          // RST/CE auf high. eingebaut in Start der Funktionen
}
*/



void rtc_write_byte(unsigned char byte) // writes byte to DS1302
                                        // veraendert RST nicht
{
   unsigned char i;

   RTC_IO_OUT();         // IO-Pin auf Ausgabe

   // bitweise schreiben: Reihenfolge D0,D1,D2,D3..D7
   for(i=0; i<8; i++){   // 8 bits ausgeben

      if(byte & 0x01)    // wenn hinterstes bit=1
                         // 1111 1111
                         // D7      D0
         RTC_IO_H();     // IO-Pin auf high
      else
         RTC_IO_L();     // IO-Pin auf lo

      // positiver CLK-Puls - 1us high, 1us low
      RTC_CLK_H();       // SCLK auf high
      delay_us(1);       // 1us muesste gehen (0.5MHz)
      RTC_CLK_L();       // SCLK auf low
      delay_us(1);       // 1us muesste gehen

      byte >>=1;         // 1 bit rechts schieben fuer naechstes bit
                         // 0111 1111 nach 1. Lauf
                         // 0011 1111 nach 2. Lauf
                         // 0001 1111 nach 3. Lauf
                         // 0000 1111 nach 4. Lauf
                         // 0000 0111 nach 5. Lauf
                         // 0000 0011 nach 6. Lauf
                         // 0000 0001 nach 7. Lauf
                         // 0000 0000 nach 8. Lauf fertig. alle 8 ausgegeben.
   }
}



unsigned char rtc_read_byte(void)  // reads ds1302 reply raw
                                   // veraendert RST nicht
{
   unsigned char i;
   unsigned char byte, temp;

   byte = 0x00;

   RTC_IO_IN();        // IO-Pin auf Input

#if DEBUG1
   outs_p(USART,PSTR("\nrtc_read_byte: ")); // Debug-Ausgabe
#endif // DEBUG

   // bits D0-D7 einlesen: Reihenfolge D0,D1,D2,D3.. D7
   for(i=0; i<8; i++){ // erst untere 4 bits D0,D1,D2,D3
                       // 0-9 1er Sekunden (00-59) bzw.
                       // 0-9 1er Minuten (00-59) bzw.
                       // 0-9 1er Stunden (00-23) bzw.
                       // 1-9 1er Datum (01-31) bzw.
                       // 1-9 1er Monat (01-12) bzw.
                       // 1-7 1er Wochentag (01-07) bzw.
                       // 0-9 1er Jahr (00-99)

                       // dann obere 4 bits D4,D5,D6,D7
                       // 0-5 10er Sekunden (00-59) bzw.
                       // 0-5 10er Minuten (00-59) bzw.
                       // 0-2 10er Stunden (00-23) bzw.
                       // 0-3 10er Datum (01-31) bzw.
                       // 0-1 10er Monat (01-12) bzw.
                       // 0-0 10er Wochentag (01-07) bzw.
                       // 0-9 10er Jahr (00-99)

         // Vorsicht mit CH-bit und 12/24-bit an Stelle D7
         // wenn diese bits gesetzt sind wird falsch gelesen !
         // daher sollten diese beiden bits ignoriert werden (0 belassen)
         // oder korrekt beruecksichtigt

#if DEBUG1
      outs_p(USART,PSTR("\ni= "));  // Debug-Ausgabe
      out_u08(USART,i,0,0);       // i ausgeben
#endif // DEBUG

      if(RTC_IO_VAL) // liest IO-Pin
         temp = 1;   // falls IO-Pin hi: 0000 0001
      else
         temp = 0;   // falls IO-Pin lo: 0000 0000

#if DEBUG1
      outs_p(USART,PSTR("\ntemp= "));  // Debug-Ausgabe
      out_u08(USART,temp,0,0);       // temp ausgeben
#endif // DEBUG

      temp <<= i;    // bit um i nach links schieben:
                     // 0000 0001  1. Lauf i=0
                     // 0000 0010  2. Lauf i=1
                     // 0000 0100  3. Lauf i=2
                     // 0000 1000  4. Lauf i=3
                     // 0001 0000  5. Lauf i=4
                     // 0010 0000  6. Lauf i=5
                     // 0100 0000  7. Lauf i=6
                     // 1000 0000  8. Lauf i=7

#if DEBUG1
      outs_p(USART,PSTR("\ntemp<<= ")); // Debug-Ausgabe
      out_u08(USART,temp,0,0);        // temp ausgeben
#endif // DEBUG

      byte |= temp;  // verodern:
                     // 0000 0001 1. Lauf
                     // 0000 0011 2. Lauf
                     // 0000 0111 3. Lauf
                     // 0000 1111 4. Lauf
                     // 0001 1111 5. Lauf
                     // 0011 1111 6. Lauf
                     // 0111 1111 7. Lauf
                     // 1111 1111 8. Lauf
                     //         D0
                     //        D1
                     //       D2
                     //      D3
                     //    D4
                     //   D5
                     //  D6
                     // D7

#if DEBUG1
      outs_p(USART,PSTR("\nbyte= "));  // Debug-Ausgabe
      out_u08(USART,byte,0,0);       // byte ausgeben
#endif // DEBUG

      // positiver CLK-Puls - 1us high, 1us low
      RTC_CLK_H();   // SCLK auf high
      delay_us(1);   // 1us (0.5MHz, damit bei 2V noch ok)
      RTC_CLK_L();   // SCLK auf low
      delay_us(1);   // 1us
   }

   // nun sind alle 8 bits im byte, D0 rechts, D7 links
   return byte;
}



unsigned char rtc_read_clock(void)  // reads ds1302 reply corrected for clock
                                    // veraendert RST nicht
{
   unsigned char byte;

   byte=rtc_read_byte();   // lese Roh-Byte

   // bei Clock vordere 4 bits * 10 + hintere 4 bits rechnen
   // untere 4 bits so nehmen wie es ist: 0000 1111
   // obere 4 bits nach rechts schieben, dann *10 und addieren

#if DEBUG
   outs_p(USART,PSTR("\nClock-Zugriff: byte= ")); // Debug-Ausgabe
   usart_putc_hex(byte);           // byte als hex ausgeben
   outs_p(USART,PSTR("->"));         // Debug-Ausgabe
   out_u08(USART,((byte & 0x0f) + ((byte >>4) * 10)),0,0); // ausgeben
#endif // DEBUG

   return ((byte & 0x0f) + ((byte >>4) * 10));
}

#endif // DS1302



void rtc(unsigned char device)  // RTC-Ausgabe auf USART/LCD/SRAM
                                // ohne Ausgabe Sekunden, Wochentag 
{
   unsigned char temp;  // RTC

   // gebe Datum/Zeit an Schreibstelle aus
   temp = rtc_read(DATE);        // Datum 3.=03
   out_u08(device,temp,2,0);     // Datum ausgeben
   
   outc(device,'.');             // . ausgeben
   temp = rtc_read(MONTH);       // Monat Sept=9
   out_u08_2(device,temp);       // Monat ausgeben
   
   outc(device,'.');             // . ausgeben
   temp = rtc_read(YEAR);        // Jahr 2016=16
   out_u08_2(device,temp);       // Jahr ausgeben
   
   outc(device,' ');             // Space ausgeben
   temp = rtc_read(HOUR);        // Stunde 11=11
   out_u08(device,temp,2,0);     // Stunde ausgeben
   
   outc(device,':');             // : ausgeben
   temp = rtc_read(MIN);         // Minuten 40=40
   out_u08_2(device,temp);       // Minuten ausgeben
   
/*
   outc(device,':');             // : ausgeben
   temp = rtc_read(SEC);         // Sekunden 00=00
   out_u08_2(device,temp);       // Sekunden ausgeben
*/   
} 



void rtc_test(void)  // RTC-Test: Ausgabe USART
{
   unsigned char sec, temp;  // RTC
   unsigned int merker=0; 
   sec = 0;                  // RTC

   // Meldung fuer RTC-Test ausgeben
   outs_p(USART,PSTR("\nRTC Test: \n"));

   // RTC stellen - stets noetig nach Batteriewechsel, sonst Probleme !
   rtc_set_time_date(          // 20.10.19 19:37:00 SO
                0x20,  // date  0x20
                0x10,  // month    0x10
                0x19,  // year        0x19
                0x19,  // hour           0x19
                0x37,  // min               0x37
                0x00,  // sec                  0x00
                0x07); // wday                    0x07

   while (1){
      temp = sec;

      sec = rtc_read(SEC);  // read seconds

      if(sec != temp){  // nur bei Aenderungen SEC diese anzeigen
         out_u08(USART,sec,3,0); 
         merker++;
         
         if (merker==10){  // nach 10s alles anzeigen
            break;         // Schleife abbrechen und anzeigen
         }
      }
   }

   // lese alle Register einzeln nacheinander aus
   outs_p(USART,PSTR("\nRTC: "));    // \nRTC: ausgeben
   temp = rtc_read(DATE);          // Datum 3.=03
   out_u08(USART,temp,2,0);        // Datum ausgeben
   outc(USART,'.');                // . ausgeben
   temp = rtc_read(MONTH);         // Monat Sept=9
   out_u08_2(USART,temp);          // Monat ausgeben
   outc(USART,'.');                // . ausgeben
   temp = rtc_read(YEAR);          // Jahr 2016=16
   out_u08_2(USART,temp);          // Jahr ausgeben
   outc(USART,' ');                // Space ausgeben
   temp = rtc_read(HOUR);          // Stunde 11=11
   out_u08(USART,temp,2,0);        // Stunde ausgeben
   outc(USART,':');                // : ausgeben
   temp = rtc_read(MIN);           // Minuten 40=40
   out_u08_2(USART,temp);          // Minuten ausgeben
   outc(USART,':');                // : ausgeben
   temp = rtc_read(SEC);           // Sekunden 00=00
   out_u08_2(USART,temp);          // Sekunden ausgeben
   outs_p(USART,PSTR(" Tag: "));
   temp = rtc_read(WDAY);          // Wochentag 6=Samstag
   out_u08(USART,temp,1,0);        // Wochentag als Zahl ausgeben
   outc(USART,'\n');               // neue Zeile ausgeben
   
   // Schlussmeldung ausgeben
   outs_p(USART,PSTR("RTC getestet\n")); // Meldung ausgeben
   
} 

